
%%
%     ---------------------------------------------------------------------
%     Covariogram-based CS data gathering & compression framework
%     ---------------------------------------------------------------------
%     Performance evaluation framework including: 
%               - generation of correlated spatio-temporal signals
%               - scheduling, sampling & data gathering
%               - signal compression (at the sensors) & reconstruction (at the sink)
%
%     Copyright (C) 2016 Mohsen Hooshmand, Davide Zordan, Michele Rossi
%     Department of Information Engineering, University of Padova, Italy
%     Email: {hooshmand, zordan, rossi}@dei.unipd.it
%     (this module was originally written by Giorgio Quer and Riccardo Masiero)
% 
%     This program is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
% 
%     This program is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
% 
%     You should have received a copy of the GNU General Public License
%     along with this program in the file "gpl-license-3.0.txt".  
%
%     If not, see <http://www.gnu.org/licenses/>.
%% 

function [x_hat, psi] = BLOCK5_sink_recovery(y, phi, training_set, x_prev, distance_mat, covraince_type, distance_vector, new_distanc_mat)
%%%% BLOCK 5 RECOVERY (by MH modified 02.11.15)
%
%
%
% DESCRIPTION: This block aims to estimate unkonwn values of sensors through basis transformation 
%              
% INPUT:  y [Mx1] = signal received by the sink
%         phi [matrix M x N] = matrix s.t. y = Phi*x
%         training_set [matrix N x L] =
%            previously recovered vectors [x^(k-L),..,x^(k-1)]
%         x_prev [Nx1] = previous value recovered for x, i.e. x^(k-1)
%         distance_mat : matrix containing distances among sensors
%         covraince_type:
%           'Simple_Cov'    -> computing psi directly form covariance matrix
%           of sensors
%           'Variogram_Est' -> computing psi form vasriogram matrix
%         distance_vector:  uniformed distance between min and max of distances
%         new_distanc_mat : New distance matrix of sensors based on
%                           the new distance_vector
%
% OUTPUT: x_hat [Nx1] = signal recovered by the sink
%         psi [NxN]= transform sparse matrix
%


M = length(y);
N = length(phi);
psi = [];

% checking special cases:
if M ==0
    % no node transmitted, previously recovered value    
    x_hat = x_prev;    
elseif M ==N
    % all nodes transmitting
    x_hat = phi'*y;
    
elseif size(training_set,2) < 2
   
    x_hat = x_prev;
    fprintf('\n Error: something went wrong in block 5 \n');
    
else
    % the standard case    
    %%%% cs-recovering
    
    % inferring the sparsification matrix Psi
    [psi] = PCA_transform_matrix(training_set,distance_mat, covraince_type, distance_vector, new_distanc_mat);
       
    
    % calculating A, s.t. y = A*s , where s is the sparse signal
    A = phi*psi;
    
    % considering the average x_m:
    x_m = mean(training_set,2);
    y_1 = y-phi*x_m;
     
    % solution of y_1 = A*s
    
    % the inner parameter of the l1-magic recovery
    inner_parameter = 1e-5;
    
    % recovering: L1 minimization by Candes
    s = SL0(A, y_1, inner_parameter);
    
    
    if isnan(s) % check
        disp('WARNING: a NaN has occured')
    end
    
    %recovered values for x (column)
    x_hat = real(x_m + psi*s);
end

end % function
