%%
%     ---------------------------------------------------------------------
%     Covariogram-based CS data gathering & compression framework
%     ---------------------------------------------------------------------
%     Performance evaluation framework including: 
%               - generation of correlated spatio-temporal signals
%               - scheduling, sampling & data gathering
%               - signal compression (at the sensors) & reconstruction (at the sink)
%
%     Copyright (C) 2016 Mohsen Hooshmand, Davide Zordan, Michele Rossi
%     Department of Information Engineering, University of Padova, Italy
%     Email: {hooshmand, zordan, rossi}@dei.unipd.it
%     (this module was originally written by Giorgio Quer and Riccardo Masiero)
% 
%     This program is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
% 
%     This program is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
% 
%     You should have received a copy of the GNU General Public License
%     along with this program in the file "gpl-license-3.0.txt".  
%
%     If not, see <http://www.gnu.org/licenses/>.
%% 

function [X_hat,tot_cost, real_error] = BLOCKmain_cs_dns...
    (X, cost_per_trasm, hops_vector, TS_Type, r_s, distance_mat, covraince_type, initial_training, time_table)
%%%% BLOCK main single signal (by MH modified 02.11.15)
% INPUT:
%         X [N * T matrix] = whole signal sensed to be analyzed, composed
%             by T (time) signals of length N
%         cost_per_trasm [N vector] =
%             cost of transmission for each node
%        hops_vector = vector where each i-th element represents the number of
%        hops among the i-th sensor and the sink x---o--o--X (= 3 hops)
%        TS_type = signals that used for the cocariance computation
%               'theoretical' -> All the X [N * T matrix]
%               'real'        -> Previous constructed values in X_hat
%        r_s =  complete smaple bitlength
%        distance_mat : matrix containing distances among sensors
%        covraince_type:
%           'Simple_Cov'    -> computing psi directly form cavriance matrix
%           of sensors
%           'Variogram_Est' -> computing psi form vasriogram matrix
%         initial_training = initial time steps that all sensors sense and
%         send thier readings
%        time_table = 0-1 matrix [N * T] that shows when each sensor have
%        to sense or not to
%
%
% OUTPUT:
%         X_hat [N x T matrix] = whole signal reconstructured, composed
%             by T (time) signals of length N
%         tot_cost = enrgy consumption for transmission of the sensed
%         readings
%         real_error[1 x T] = vector with all the real errors of the net
%




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% initialization
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

NoS=size(X,1);

X_hat = zeros(size(X));
x_hat = [];
[~,T] = size(X);


real_error = zeros(1 , T);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%ITERATIVE TECHNIQUE:
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
[ distance_vector , new_distances ] = distance_reducing( distance_mat);


    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % actual simulation: initial training phase
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    for k = 1: initial_training;        
        x = X(:,k);     % signal transmitted at time k
        
        [y,phi] = BLOCK4_network_tdma(x,time_table,k);
        
        % at the sink:
        [x_hat,~] = BLOCK5_sink_recovery(y,phi,[],[], distance_mat, covraince_type, distance_vector, new_distances);
        
        
        X_hat(:,k) = x_hat;
        
        real_error(k) = mean(sqrt((x_hat - x).^2));
        
    end
    trans_bit = cost_per_trasm.*initial_training;
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % actual simulation: here it starts the CS+PCA
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    

    for k = initial_training+1:T

        x_prev = x_hat;
        
        x = X(:,k);     % signal transmitted at time k
        
        [y,phi] = BLOCK4_network_tdma(x,time_table,k);
        
        % at the sink:
        switch(TS_Type)
            case 'theoretical'
                training_set = X; 
            case 'real'
                training_set = X_hat(:,1:k-1);             
            otherwise
                training_set = X;
                display('Theoretical TS selected by default');
        end
        [x_hat,~] = BLOCK5_sink_recovery(y,phi,training_set,x_prev, distance_mat, covraince_type, distance_vector, new_distances);
        
        % saving results
        X_hat(:,k) = x_hat;
        
        trans_temp=time_table(:,k);
        
        trans_bit = trans_bit + trans_temp.*r_s;
        

        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % part of performance analysis
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        % calculation of the real error        
       real_error(k) = mean(sqrt((x_hat - x).^2));
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                 
    end
    tot_cost=energy_consumption_header( hops_vector, [1:NoS], trans_bit );   
end
