%%
%     ---------------------------------------------------------------------
%     Covariogram-based CS data gathering & compression framework
%     ---------------------------------------------------------------------
%     Performance evaluation framework including: 
%               - generation of correlated spatio-temporal signals
%               - scheduling, sampling & data gathering
%               - signal compression (at the sensors) & reconstruction (at the sink)
%
%     Copyright (C) 2016 Mohsen Hooshmand, Davide Zordan, Michele Rossi
%     Department of Information Engineering, University of Padova, Italy
%     Email: {hooshmand, zordan, rossi}@dei.unipd.it
%     (this module was originally written by Giorgio Quer and Riccardo Masiero)
% 
%     This program is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
% 
%     This program is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
% 
%     You should have received a copy of the GNU General Public License
%     along with this program in the file "gpl-license-3.0.txt".  
%
%     If not, see <http://www.gnu.org/licenses/>.
%% 
    
function [X_hat,tot_cost, real_error] = BLOCKmain_cs_rns...
    (X, cost_per_trasm, ptx_parameters, technique, hops_vector, TS_Type)
%%%% BLOCK main single signal (by GQ&RM modified 21.10.09)
%%%% BLOCK main single signal (by MH remodified 02.11.15)
% INPUT:
%         X [N * T matrix] = whole signal sensed to be analyzed, composed
%             by T (time) signals of length N
%         cost_per_trasm [N vector] =
%             cost of transmission for each node
%         iterative_parameters [5x1 vector]=
%           error_type [integer] = 1,2,3
%               1: ||y-phi_prev*x_prev||_2 / ||y||_2
%               2: ||y_prev - phi*x||_2 / ||y_prev||_2
%               3: || [y;y_prev] - [phi_prev*x_prev, phi*x||_2 /
%                           |[y_prev ; y]||_2
%           epsilon = threshold under which the error is negligible
%           aimd_add_decr = average n of nodes that should stop transmitting
%               if the error is under threshold
%           aimd_mult_incr = coefficient to multiply the p_tx
%           initial_p_tx = initial value for the p_tx
%         globecom_parameters [x1 vector] =
%           K = training phase length
%           zeta = ratio between monitoring and training phase
%           p_tx = probability of transmitting in the monitorning phase
%         technique =
%               1 -> Globecom technique
%               2 -> iterative technique with fixed p_tx
%               3 -> iterative technique, full behaviour
%         hop_vector [N 1] = Numebr of hops from each sensor to the sink
%         TS_type = signals that used for the cocariance computation
%               'theoretical' -> All the X [N * T matrix]
%               'real'        -> Previous constructed values in X_hat
%
%
% OUTPUT:
%         X_hat [N x T matrix] = whole signal reconstructured, composed
%             by T (time) signals of length N
%         tot_cost = enrgy consumption for transmission of the sensed
%         readings
%         real_error[1 x T] = vector with all the real errors of the net
%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% initialization
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
X_hat = zeros(size(X));
x_hat = [];
[~,T] = size(X);

tot_cost = zeros(1, T);
real_error = zeros(1 , T);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%ITERATIVE TECHNIQUE:
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if technique > 1
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % parameters
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    initial_training = 2;  % length of the initial training
    
    % Feedback parameters
    % error_type = iterative_parameters(1);
    % epsilon = iterative_parameters(2);
    % aimd_add_decr = iterative_parameters(3);
    % aimd_mult_incr = iterative_parameters(4);
    initial_p_tx = ptx_parameters;
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % actual simulation: initial training phase
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    for k = 1: initial_training;
        p_tx = 1;       % all nodes are transmitting
        x = X(:,k);     % signal transmitted at time k
        % in the network:
        [y,phi, tx_nodes_indexes] = BLOCK4_network_cs_rns(x, p_tx);
        
        % at the sink:
        [x_hat] = BLOCK5_sink_recovery_cs_rns(y,phi,[],[]);
        
        % I write the results
        X_hat(:,k) = x_hat;
        
        tot_cost(k) = energy_consumption_header(hops_vector, tx_nodes_indexes,cost_per_trasm);
        
        real_error(k) = mean(sqrt((x_hat - x).^2));
        
    end;
    %     trans_bit = cost_per_trasm.*initial_training;
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % actual simulation: here it starts the CS+PCA
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    
    p_tx = initial_p_tx;
    training_set_dim = 2;
    % normal algorithm behaviour
    for k = initial_training+1:T
        x_prev = x_hat;
        x = X(:,k);     % signal transmitted at time k
        
        % in the network:
        [y,phi, tx_nodes_indexes] = BLOCK4_network_cs_rns(x, p_tx);
        
        % at the sink:
        switch(TS_Type)
            case 'theoretical'
                training_set = X;
            case 'real'
                training_set = X_hat(:,k-training_set_dim:k-1);
            otherwise
                training_set = X;
                display('Theoretical TS selected by default');
        end
        [x_hat] = BLOCK5_sink_recovery_cs_rns(y,phi,training_set,x_prev);
        
        X_hat(:,k) = x_hat;
        
        % total cost (for the network)
        tot_cost(k) = energy_consumption_header(hops_vector, tx_nodes_indexes, cost_per_trasm);
        
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % part of performance analysis
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        % calculation of the real error
        real_error(k) = mean(sqrt((x_hat - x).^2));
    end;
end;
end % function
