%%
%     ---------------------------------------------------------------------
%     Covariogram-based CS data gathering & compression framework
%     ---------------------------------------------------------------------
%     Performance evaluation framework including: 
%               - generation of correlated spatio-temporal signals
%               - scheduling, sampling & data gathering
%               - signal compression (at the sensors) & reconstruction (at the sink)
%
%     Copyright (C) 2016 Mohsen Hooshmand, Davide Zordan, Michele Rossi
%     Department of Information Engineering, University of Padova, Italy
%     Email: {hooshmand, zordan, rossi}@dei.unipd.it
% 
%     This program is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
% 
%     This program is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
% 
%     You should have received a copy of the GNU General Public License
%     along with this program in the file "gpl-license-3.0.txt".  
%
%     If not, see <http://www.gnu.org/licenses/>.
%% 
    

function [U] = PCA_transform_matrix(set_signals,distance_mat, covraince_type,distance_vector, new_distanc_mat)
%% CS meets PCA using simple cov or - BLOCK2 PCA_transform_matrix (by MH modified 02.11.15)

% DESCRIPTION: This block generates the PCA basis for a given set
%              of signals. From the input signals, Sigma will be computed 
%              from coriance among sensor readings. Then, based on covariance type,
%              the simple covariance computation or variogram-based method will
%              generate the sparse basis (U).
%
%
% INPUT:  set_signals: a matrix of order Nxn, made of n 1-D Nx1 (columns)
%         signals
%         distance_mat : matrix containing distances among sensors
%         covraince_type:
%           'Simple_Cov'    -> computing psi directly form cavriance matrix
%           of sensors
%           'Variogram_Est' -> computing psi form variogram matrix
%         distance_vector:  uniformed distance between min and max of distances
%         new_distanc_mat : New distance matrix of sensors based on
%                           the new distance_vector
%
%
% OUTPUT: U, the N x N matrix whose columns are the eigenvectors of Sigma
%            ordered according to the corresponding eigenvalues' decreasing
%            order. Sigma is the signal correlation matrix.
%


switch(covraince_type)
    case 'Simple_Cov'
        Sigma = cov(set_signals');
        [ U ] = direct_eig_computation( Sigma );
    case 'Variogram_Est'
        Sigma = covriance_estimation_variogram (set_signals, distance_mat,distance_vector, new_distanc_mat);
        [ U ] = direct_eig_computation( Sigma );
    otherwise
        display('Default: covriance reconstruction using direct and simple method');
        Sigma = cov(set_signals');
        [ U ] = direct_eig_computation( Sigma );
end
end
