%%
%     ---------------------------------------------------------------------
%     Covariogram-based CS data gathering & compression framework
%     ---------------------------------------------------------------------
%     Performance evaluation framework including: 
%               - generation of correlated spatio-temporal signals
%               - scheduling, sampling & data gathering
%               - signal compression (at the sensors) & reconstruction (at the sink)
%
%     Copyright (C) 2016 Mohsen Hooshmand, Davide Zordan, Michele Rossi
%     Department of Information Engineering, University of Padova, Italy
%     Email: {hooshmand, zordan, rossi}@dei.unipd.it
%     (this module was originally written by Giorgio Quer and Riccardo Masiero)
% 
%     This program is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
% 
%     This program is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
% 
%     You should have received a copy of the GNU General Public License
%     along with this program in the file "gpl-license-3.0.txt".  
%
%     If not, see <http://www.gnu.org/licenses/>.
%% 
    

function [U,m] = PCA_transform_matrix_cs_rns(set_signals)
%% CS meets PCA - BLOCK2 PCA_transform_matrix (by R&G, 25/02/2009)

% DESCRIPTION: This block aims to generate the PCA basis for a given set
%              of signals. From the input signals, their mean and 
%              covariance matrix Sigma are computed. Then, the eigenvalues 
%              and eigenvectors of Sigma are computed and ordered according
%              to the eigenvalues' decreasing order. The output matrix U is
%              the matrix whose columns are the eigenvectors of Sigma in
%              such order.
% INPUT:  set_signals, a matrix of order Nxn, made of n 1-D Nx1 (columns) 
%         signals
% OUTPUT: U, the N x N matrix whose columns are the eigenvectors of Sigma
%            ordered according to the corresponding eigenvalues' decreasing 
%            order. Sigma is the signal correlation matrix.
%         m, the Nx1 mean vector of the signal set
m = mean(set_signals')'; % mean of the signal set Nx1

Sigma = cov(set_signals'); % covariance matrix of the signal set NxN

[T,E] = eig(Sigma); % eigenvectors and eigenvalues of Sigma

Td = double(T); % eigenvectors of Sigma

Ed = double(E); % diag(Ed) = eigenvalues of Sigma

[~, permutazione] = sort(diag(Ed), 'descend'); % eigenvalues of Sigma 
                                                 % in descending order
                                                   
U = permuta(Td,permutazione); % eigenvectors in descending order

%% function permuta
%INPUT:  matrix, a matrix to order according to a given sort vector
%        permutazione, the sort vector
%OUTPUT: matrix_p, the sorted matrix
function [matrix_p] = permuta(matrix,permutazione)

for p = 1:length(permutazione)   
  matrix_p(:,p) = matrix(:,permutazione(p));     
end



