/*
*     ---------------------------------------------------------------------
*     Covariogram-based CS data gathering & compression framework
*     ---------------------------------------------------------------------
*     Performance evaluation framework including:
*               - generation of correlated spatio-temporal signals
*               - scheduling, sampling & data gathering
*               - signal compression (at the sensors) & reconstruction (at the sink)
*
*     Copyright (C) 2016 Mohsen Hooshmand, Davide Zordan, Michele Rossi
*     Department of Information Engineering, University of Padova, Italy
*     Email: {hooshmand, zordan, rossi}@dei.unipd.it
*
*     This program is free software: you can redistribute it and/or modify
*     it under the terms of the GNU General Public License as published by
*     the Free Software Foundation, either version 3 of the License, or
*     (at your option) any later version.
*
*     This program is distributed in the hope that it will be useful,
*     but WITHOUT ANY WARRANTY; without even the implied warranty of
*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*     GNU General Public License for more details.
*
*     You should have received a copy of the GNU General Public License
*     along with this program in the file "gpl-license-3.0.txt".
*
*     If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <math.h>
#include "brentFitCpp.h"
#include "mex.h"

// ----------------------------------------------------------------------------------------
// NOTE:
// The following routines have been taken from the book:
// Numerical Recipes in C: "The Art of Scientific Computing", 2nd Edition. Cambridge, 1992.
// ----------------------------------------------------------------------------------------

#define GOLD 1.618034
#define GLIMIT 100.0
#define TINY 1.0e-20
#define SHFT(a,b,c,d) (a)=(b);(b)=(c);(c)=(d);
// Here GOLD is the default ratio by which successive intervals are magnified;
// GLIMIT is the maximum magnification allowed for a parabolic-fit step.

#define ITMAX 1000
#define CGOLD 0.3819660
#define ZEPS 1.0e-10
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))

using namespace std;


// MAIN ROUTINE
// res=brentFitCpp(a,c,delta,distances,corValues,distLng,scalar,flag);
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    // prepare variables
    double a,b,c,delta,scalar;
    int distLng, flag;
    double *output = new double[2];
    double *distances;
    double *corValue;
    
    // get input
    if( nrhs!=8)
        cout << "Error: number of input elements doesn't match!";
    else
    {
        a=mxGetScalar(prhs[0]);
        c=mxGetScalar(prhs[1]);
        delta=mxGetScalar(prhs[2]);
        distances=mxGetPr(prhs[3]);
        corValue=mxGetPr(prhs[4]);
        distLng=(int) mxGetScalar(prhs[5]);
        scalar=mxGetScalar(prhs[6]);
        flag=(int) mxGetScalar(prhs[7]);
    }
    
	b=( (c-a)*CGOLD + a);
    // prepare output
    plhs[0]=mxCreateDoubleMatrix(1, 2, mxREAL);
    output=mxGetPr(plhs[0]);

    brent( a, b, c, delta, distLng, distances, corValue, output, scalar , flag);
    
    return;
}

// SEARCHING FOR THE MINIMUM IN AN INTERVAL
// Given a function f, and given a bracketing triplet of abscissas ax, bx, cx (such that bx is between ax and cx, and f(bx) is
// less than both f(ax) and f(cx)), this routine isolates the minimum to a fractional precision of about tol using Brent’s method.
// The abscissa of the minimum is returned as xmin, and the minimum function value is returned as brent, the returned function value.
void brent(double ax, double bx, double cx, double tol, int distLng, double distances[], double corValue[], double output[], double scalar, int flag)
{
	int iter;
	double a,b,d,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm;
	double e=0.0;											// This will be the distance moved on the step before last.
	
	a=(ax < cx ? ax : cx);									// a and b must be in ascending order, but input abscissas need not be.
	b=(ax > cx ? ax : cx);
	
	x=w=v=bx;
	//fw=fv=fx=(*f)(x);										// Initializations, we take f in bx (end on the interval)

    if(flag!=0)
        fw=fv=fx=mseComputation( x, scalar, distances, corValue, distLng);
    else 
        fw=fv=fx=mseComputation( scalar, x, distances, corValue, distLng);	
	
	for (iter=1;iter<=ITMAX;iter++) {						// Main program loop.
		xm=0.5*(a+b);
		
		tol2=2.0*(tol1=tol*fabs(x)+ZEPS);
		
		if (fabs(x-xm) <= (tol2-0.5*(b-a))) {				// Test for done here.
            output[0]=x;																		// never get here
            output[1]=fx;
            return;
		}
		
		if (fabs(e) > tol1) {								// Construct a trial parabolic fit.
	
			r=(x-w)*(fx-fv);
			q=(x-v)*(fx-fw);
			p=(x-v)*q-(x-w)*r;
			q=2.0*(q-r);
			if (q > 0.0) p = -p;
			q=fabs(q);
			etemp=e;
			e=d;
			
			if (fabs(p) >= fabs(0.5*q*etemp) || p <= q*(a-x) || p >= q*(b-x))		// These conditions determine the acceptability of the parabolic fit.
			
				d = CGOLD*(e=(x >= xm ? a-x : b-x));								// Here we take the golden section step into the larger of the two segments.
			
			else {
			
				d=p/q;																// Take the parabolic step.
				u=x+d;
				if (u-a < tol2 || b-u < tol2)
					d=SIGN(tol1,xm-x);
			}
		} else {
			
			d=CGOLD*(e=(x >= xm ? a-x : b-x));
		
		}
		
		u=(fabs(d) >=tol1 ? x+d : x+SIGN(tol1,d));
		
        if(flag!=0)
            fu=mseComputation( u, scalar, distances, corValue, distLng);
        else 
            fu=mseComputation( scalar, u, distances, corValue, distLng);	
		
		if (fu <= fx) {																// Now decide what to do with our function evaluation.
			
			if (u >= x) a=x; else b=x;												// Housekeeping follows:
			SHFT(v,w,x,u)
			SHFT(fv,fw,fx,fu)
			
		} else {
			if (u < x) a=u; else b=u; if (fu <= fw || w == x) {
				v=w;
				w=u;
				fv=fw;
				fw=fu;
			} else if (fu <= fv || v == x || v == w) {
				v=u;
				fv=fu;
			}
		}
	}																				// Done with housekeeping. Back for another iteration.
	
	cout << endl << "WARNING: too many iterations in brent (FuncSearch.c)" << endl << endl;
	// Barbatrucco! Funziona ma tecnicamente è sbagliato
    if(flag!=0)
    {
        output[0]=1;
        output[1]=mseComputation( 1, scalar, distances, corValue, distLng);
    }
    else 
    {
        output[0]=ax;
        output[1]=mseComputation( scalar, ax, distances, corValue, distLng);
    }
}

double mseComputation(double a, double b, double distances[], double corValue[], int distLng)
{
    double mse=0;
    for(int i=0; i<distLng; i++)
        mse+=pow( ( a*exp(b*distances[i]) - corValue[i] ), 2.0 );
    return mse;
}
