%%
%     ---------------------------------------------------------------------
%     Covariogram-based CS data gathering & compression framework
%     ---------------------------------------------------------------------
%     Performance evaluation framework including: 
%               - generation of correlated spatio-temporal signals
%               - scheduling, sampling & data gathering
%               - signal compression (at the sensors) & reconstruction (at the sink)
%
%     Copyright (C) 2016 Mohsen Hooshmand, Davide Zordan, Michele Rossi
%     Department of Information Engineering, University of Padova, Italy
%     Email: {hooshmand, zordan, rossi}@dei.unipd.it
% 
%     This program is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
% 
%     This program is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
% 
%     You should have received a copy of the GNU General Public License
%     along with this program in the file "gpl-license-3.0.txt".  
%
%     If not, see <http://www.gnu.org/licenses/>.
%% 
    
function [Sigma] = covriance_estimation_variogram (set_signals, distance_mat, distance_vector, new_distanc_mat)
%% covarince computation improvement by variogram (by MH modified 02.11.15)

% DESCRIPTION: This block estimates the covariance matrix \Sigma through the covariogram-based
%               approach of Section VII of the following paper:
%
%               M. Hooshmand, M. Rossi, D. Zordan, M. Zorzi, Covariogram-based Compressive Sensing
%               for Environmental Wireless Sensor Networks, IEEE Sensors, 2016.
%
% INPUT:  
%         set_signals: a matrix of order Nxn, made of n 1-D Nx1 (columns)
%         signals
%         distance_mat : matrix containing distances among sensors
%         distance_vector:  uniformed distance between min and max of distances
%         new_distanc_mat : New distance matrix of sensors based on
%                           the new distance_vector
%
% OUTPUT: 
%         Sigma, the N x N covariance (basis) matrix.
%
NoS=length(distance_mat);
cov_matrix=cov(set_signals');
variances=diag(cov_matrix);
std_dev=sqrt(variances);

cov_function =zeros(size(cov_matrix));
for jj=1:NoS
    for ii=1:NoS
        cov_function(ii,jj)=cov_matrix(ii,jj)/std_dev(jj)/std_dev(ii);
    end
end

new_distances_vec = reshape(new_distanc_mat, NoS*NoS,1);
cov_function_vec = reshape(cov_function, NoS*NoS,1);
discount=ones(length(distance_vector),1);
general_cov_function=zeros(length(distance_vector),1);
for i= 1:length(distance_vector)
    tmp=find(new_distances_vec==distance_vector(i));
    discount(i)=length(tmp)/length(new_distances_vec);
    if tmp
        general_cov_function(i)=mean(cov_function_vec(tmp)) ;
    end
end

               
% 2 VARIABLES BRENT CPP IMPLEMENTED METHOD ____________________
% Parameters needed
a=-3;
b=2;
delta=1e-6;
threshold=1e-3;
% While loop controller
cMaxCtrl=30;
ctrl=0;
% Initialization
aOpt=1;
% First Brent iteration
res=brentFitCpp(a,b,delta,distance_vector,general_cov_function,length(distance_vector),aOpt,0);
bOld=res(1);
res=brentFitCpp(a,b,delta,distance_vector,general_cov_function,length(distance_vector),bOld,1);
aOld=res(1);
aOpt=res(1);
% While loop until difference between variables is under a certain threshold
flag=1;
while flag && ctrl<cMaxCtrl
    ctrl=ctrl+1;
    res=brentFitCpp(a,b,delta,distance_vector,general_cov_function,length(distance_vector),aOpt,0);
    bOpt=res(1);
    res=brentFitCpp(a,b,delta,distance_vector,general_cov_function,length(distance_vector),bOpt,1);
    aOpt=res(1);
    if( sqrt( (bOpt-bOld)^2 + (aOpt-aOld)^2) < threshold )
        flag=0;
    else
        bOld=bOpt;
        aOld=aOpt;
    end
end
Sigma=sigmaBuildCppVar(NoS,distance_mat,length(distance_vector),distance_vector,[aOpt, bOpt],std_dev);

end
