
%%
%     ---------------------------------------------------------------------
%     Covariogram-based CS data gathering & compression framework
%     ---------------------------------------------------------------------
%     Performance evaluation framework including: 
%               - generation of correlated spatio-temporal signals
%               - scheduling, sampling & data gathering
%               - signal compression (at the sensors) & reconstruction (at the sink)
%
%     Copyright (C) 2016 Mohsen Hooshmand, Davide Zordan, Michele Rossi
%     Department of Information Engineering, University of Padova, Italy
%     Email: {hooshmand, zordan, rossi}@dei.unipd.it
% 
%     This program is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
% 
%     This program is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
% 
%     You should have received a copy of the GNU General Public License
%     along with this program in the file "gpl-license-3.0.txt".  
%
%     If not, see <http://www.gnu.org/licenses/>.
%% 
    

function [synt_data, xc , yc, RoT] = generate_synthetic_signal(dim_grid, dim_time, C, r)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%GENERATE_SYNTHETIC_SIGNAL
% generates a synthetic signal according to the given spatial correlation matrix 
% and temporal correlation coefficient
%
% Inputs:
%   dim_grid                the spatial dimension of the output signal
%   dim_time                the number of time instants to be generated
%   C [dim_gridxdim_grid]   the spatial correlation matrix
%   r                       the temporal correlation coefficient
%
% Outputs:
%   synt_data               signal generated
%    [dim_gridxdim_gridxdim_time]
%   xc, yc                  spatial correlation of the synthetic signal
%   RoT                     temporal correlation coefficient of the
%                           synthetic signal
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% synthetic signal is generated on a square grid 
dim_lato_rete = dim_grid;
% coefficient AR process; time evolution of the loss field 
a_time = r;             
% standard deviation of the noise in the AR process
b_time = sqrt(1 - a_time^2);        
T = dim_time;
miu = 0;
sigma = 1;

R = (1/dim_lato_rete).*fft2(C);
Z_tot = zeros(dim_lato_rete,dim_lato_rete,T);

for i = 1:T  % each time instant 
    
    if i == 1
        Z_iid = miu + sigma.*randn(dim_lato_rete);
    else
        Z_iid = a_time.*real(Z_iid_t1) + b_time.*randn(dim_lato_rete);
    end
    S_Lf_iid = (1/dim_lato_rete).*(fft2(Z_iid));
%     S_Lf = (R.^(1/2)) .*S_Lf_iid;
    S_Lf = sqrt(R).*S_Lf_iid;
    Z = (dim_lato_rete).*(ifft2(S_Lf));
    Z = real(Z);
    
    Z_iid_t1 = Z_iid;
    
    Z_tot(:,:,i) = Z;
end


RoT = 0;

xc = 0;
yc = 0;
synt_data = Z_tot;

end
