%%
%     ---------------------------------------------------------------------
%     Covariogram-based CS data gathering & compression framework
%     ---------------------------------------------------------------------
%     Performance evaluation framework including:
%               - generation of correlated spatio-temporal signals
%               - scheduling, sampling & data gathering
%               - signal compression (at the sensors) & reconstruction (at the sink)
%
%     Copyright (C) 2016 Mohsen Hooshmand, Davide Zordan, Michele Rossi
%     Department of Information Engineering, University of Padova, Italy
%     Email: {hooshmand, zordan, rossi}@dei.unipd.it
%
%     This program is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
%
%     This program is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
%
%     You should have received a copy of the GNU General Public License
%     along with this program in the file "gpl-license-3.0.txt".
%
%     If not, see <http://www.gnu.org/licenses/>.
%%

 function [model n n_op] = ltc_encoder(x,tol)

% -------------------------------------------------------------------------
% LTC
% =========================================================================
% LightWeight Temporal Compression
%
% Input:   
% -------------------------------------------------------------------------
% x        Row Vector Samples 
%
% tol      Maximun Allowed Deviation Per Sample (Tolerance)
%
% Output:   
% ----------------
% model    Set of Segments with the following format:
% 
%        2 Cols: Points Defining Segments. 
%                Each Segment is composed by 2 Sucessive Points 
%                PA=(x(i),y(i)) PB=(x(i+1),y(i+1)) 
%        n Rows: One For The Start Point of Each Segment + Last point
%        
%        X(1)  Y(1)   
%        X(2)  Y(2)
%        ...   ...  
%        X(n)  Y(n)
%
% n_op     a vector containing the number of operations
%          [c_div, c_mul, c_sum, c_cmp]


% Configuration

%   PLOT=0;
%   
%   DISPLAY_OPS=0;
  
  N=length(x);

  % Ops.Counters
  c_div=0;
  c_mul=0;
  c_sum=0;
  c_cmp=0;

  %flag=0;
  %j=0;

  index_ru=1;
  index_rl=2;
  index_lu=3;
  index_ll=4;
  
  flag=0;
  
  % ----------------------------------
  % First Point of DataSet

  n=1;
  model=zeros(N,2);
  
  % Limits
  limit=zeros(N,4);

  i=1;
  j=1;
  
  z=x(1);
  
  ru=z;
  rl=z;
      
  lu=z;
  ll=z;
  
  limit(i,index_ru)=ru;
  limit(i,index_rl)=rl;
  limit(i,index_lu)=lu;
  limit(i,index_ll)=ll;
 
  while(i<N)
  
      % First Point
      %-------------
      %z=(limit(i,index_lu)+limit(i,index_ll))/2;
      z=(lu+ll)/2;
      
      c_div=c_div+1;
      
      model(n,1)=i;
      model(n,2)=z;
      n=n+1;
        
      % Second Point
      %-------------
      
      j=i+1;
      
      r=x(j); % Set Middle Point
      
      ru=r+tol;
      rl=r-tol;
      
      lu=r+tol;
      ll=r-tol;
      
      c_sum=c_sum+1;
      
      limit(j,index_ru)=ru;
      limit(j,index_rl)=rl;
      
      limit(j,index_lu)=lu;
      limit(j,index_ll)=ll;
  
      flag=0;
      
      while(flag==0 && j<N)
        
        % 2 Calculate HighLine Connecting z and UL
        % 3 Calculate LowLine Connecting z and LL
        
        lu_slope=(lu-z)/(j-i);       
        ll_slope=(ll-z)/(j-i);             
        
        c_div=c_div+2;
        
        % Next Point for High/low Lines
        lu_next=lu+lu_slope;
        ll_next=ll+ll_slope;
        
        c_sum=c_sum+2;
        
        % 4 Get Next Data Point       
        r=x(j+1);
  
        ru_next=r+tol;
        rl_next=r-tol;
        
        c_sum=c_sum+2;
        
        % 5: End IF:
        % HighLine Above rl
        % LowLine Below ru
        
        if(lu_next<rl_next)
            flag=1;
        end;
        
        if(ll_next>ru_next)
            flag=1;
        end;
        
        c_cmp=c_cmp+2;
        
        if(flag==0)
        
            % Point Falls on the Limits.
            
            if(lu_next>ru_next)
                lu=ru_next;
            else
                lu=lu_next;    
            end;
            
            if(ll_next<rl_next)
                ll=rl_next;
            else
                ll=ll_next;    
            end;
            
            c_cmp=c_cmp+2;
            
            j=j+1;  
            
            limit(j,index_ru)=ru;
            limit(j,index_rl)=rl;
            
            limit(j,index_lu)=lu;
            limit(j,index_ll)=ll;

        end;     
       
      end;    

      i=j;
      
  end;   

  % 9 Output LAST Z.
  model(n,1)=N;
  %model(n,2)=(limit(i,index_lu)+limit(i,index_ll))/2;;
  model(n,2)=(lu+ll)/2;
  
  % Cut Model
  model=model(1:n,[1 2]);

  n_op=[c_div,c_mul,c_sum,c_cmp];   
  
  
%   if DISPLAY_OPS==1
%     fprintf('\n   >>>LTC Model(%d) Elaboration:',n);
%     fprintf('\n   >>>%6d %6d %6d %6d',c_div,c_mul,c_sum,c_cmp);
%     fprintf('\n');
%   end;
  
  
  % Model Plot
  
%   if PLOT==1
%       
%       figure;
%       plot(1:N,x,'r.');
%       
%       hold on;
%       for i=1:(n-1)
%           x1=model(i,1);
%           x2=model(i+1,1);
%           y1=model(i,2);
%           y2=model(i+1,2);
%           line([x1 x2] , [y1 y2], 'Color','k','LineStyle','-')
%       end
%       hold off;
%   end;    
  

end  

