%%
%     ---------------------------------------------------------------------
%     Covariogram-based CS data gathering & compression framework
%     ---------------------------------------------------------------------
%     Performance evaluation framework including:
%               - generation of correlated spatio-temporal signals
%               - scheduling, sampling & data gathering
%               - signal compression (at the sensors) & reconstruction (at the sink)
%
%     Copyright (C) 2016 Mohsen Hooshmand, Davide Zordan, Michele Rossi
%     Department of Information Engineering, University of Padova, Italy
%     Email: {hooshmand, zordan, rossi}@dei.unipd.it
%
%     This program is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
%
%     This program is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
%
%     You should have received a copy of the GNU General Public License
%     along with this program in the file "gpl-license-3.0.txt".
%
%     If not, see <http://www.gnu.org/licenses/>.
%%

clear all
close all
clc;

mex -g brentFitCpp.cpp
mex -g sigmaBuildCppVar.cpp
mex -g sigmaBuildCppWWV.cpp

warning off;

%%
% Initializations start here -------------------------->>>
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

T = 100;            % time instants in each data collection round
tplMax = 20;        % Number of different topologies (at least must be two)

%%
%Trasmission Range of each Sensor
TR=25; % radius of transmission Range

%%
% Maximum Distance of each Cluster
MD=50;

%%
%Number of Nodes in the simulation Field
NoS=50;

%%
% Size of the simulation Field
SoF_X=100;
SoF_Y=100;

%%
% x and y Coordinates of the Sink node
% (default is in the center of the area)
sink.x=0.5*SoF_X;
sink.y=0.5*SoF_Y;

%%
% number of bits per collected sample
r_s=16; % complete smaple bitlength

%% DSC parameters
rate4=4; % bin bitlength
rate8=8; % bin bitlength
all_indeces4 = zeros(2^r_s,1); % compressed data for bin-lenth = 4 bits
all_indeces8 = zeros(2^r_s,1); % compressed data for bin-lenth = 8 bits
for i=1:2^r_s
    all_indeces4(i) = rem(i-1,2^rate4);
    all_indeces8(i) = rem(i-1,2^rate8);
end

%%
% correlation vector - each element is (rho_{space}, rho_{time})
% first and second columns respectively contain the temporal and spatial correlation
% simulation results are obtained for all the pairs in this vector
correlations=[0.8 0.1; 0.5 1; 0.5 5; 0 5];

lentgh_r=size(correlations,2); % # of correlations pairs
range = 10; % split probability range [0,1] in "range" intervals

%%
% Next we initialize matrices, for each algo, which will be later used to save the results
% the name of the compression schemes is the same used in the following paper:
%
% Mohsen Hooshmand, Michele Rossi, Davide Zordan, Michele Zorzi,
% "Covariogram-based Compressive Sensing for Environmental Wireless Sensor
% Networks," IEEE Sensors, 2016.
%$

% algo name: CS-lower-bound
report_CS_lower_bnd = zeros(tplMax,range,2);
error_CS_lower_bnd = zeros(lentgh_r,range);
energy_CS_lower_bnd = zeros(lentgh_r,range);
tx_energy_CS_lower_bnd =zeros(tplMax,range);

% algo name: CS-RNS
report_CS_rns = zeros(tplMax,range,2);
error_CS_rns = zeros(lentgh_r,range);
energy_CS_rns = zeros(lentgh_r,range);
tx_energy_CS_rns = zeros(tplMax,range);

% algo name: CS-DNS
report_CS_dns = zeros(tplMax,range,2);
error_CS_dns = zeros(lentgh_r,range);
energy_CS_dns = zeros(lentgh_r,range);
tx_energy_CS_dns = zeros(tplMax,range);

% algo name: KRON-CS
report_KRON_CS = zeros(tplMax,range,2);
error_KRON_CS = zeros(lentgh_r,range);
energy_KRON_CS = zeros(lentgh_r,range);
tx_energy_KRON_CS = zeros(tplMax,range);

% algo name: CB-CS (iid)
report_CB_CS_iid = zeros(tplMax,range,2);
error_CB_CS_iid = zeros(lentgh_r,range);
energy_CB_CS_iid = zeros(lentgh_r,range);
tx_energy_CB_CS_iid = zeros(tplMax,range);

% algo name: CB-CS (dct)
report_CB_CS_dct = zeros(tplMax,range,2);
error_CB_CS_dct = zeros(lentgh_r,range);
energy_CB_CS_dct = zeros(lentgh_r,range);
tx_energy_CB_CS_dct = zeros(tplMax,range);

% algo name: DSC
report_dsc = zeros(tplMax,range,2);
error_dsc = zeros(lentgh_r,range);
energy_dsc = zeros(lentgh_r,range);
tx_energy_dsc = zeros(tplMax,range);

% algo name: DCT
error_dct = zeros(lentgh_r,range);
energy_dct = zeros(lentgh_r,range);
tx_energy_dct =zeros(tplMax,range);
mean_cr_dct = zeros(tplMax,range);
mean_rmse_dct = zeros(tplMax,range);
mean_energy_dct = zeros(tplMax,range);
mean_max_error_dct = zeros(tplMax,range);

% algo name: LTC
error_ltc = zeros(lentgh_r,range);
energy_ltc = zeros(lentgh_r,range);
tx_energy_ltc =zeros(tplMax,range);
mean_cr_ltc = zeros(tplMax,range);
mean_rmse_ltc = zeros(tplMax,range);
mean_energy_ltc = zeros(tplMax,range);
mean_max_error_ltc = zeros(tplMax,range);

% The following variables contain the number of MCU cycles for each
% operation. The reference architecture is the Texas Instruments MSP430
c_per_div = 405;
c_per_mul = 395; % 153 if performed with hardware
c_per_sum = 184;
c_per_comp = 37;

op_cost = [c_per_div; c_per_mul; c_per_sum; c_per_comp];

% energy expenditure per MCU cycle
energy_per_instruction = 0.726e-9; %0.726 nJ

dim_grid = 500;     % grid dimension for the 2D signal (500 by 500 points evenly spaced in the simulation area)
corr_model ='b';    % correlation model type

% 2D signal Grid
x = -(ceil(dim_grid/2) - 1):1:floor(dim_grid/2);
y = -(ceil(dim_grid/2) - 1):1:floor(dim_grid/2);
[X,Y] = meshgrid(x,y);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CORRELATION MATRIX GENERATOR
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Coefficient for PE-RQ Signal Generating Models
alpha = sqrt(2*(sqrt(dim_grid)-1)^2); % Scaling factor %

for r_id=1:size(correlations,1)
    disp(['Correlations index = ', num2str(r_id),'/', num2str(length(correlations))]);
    r=correlations(r_id,1);
    ros=correlations(r_id,2);
    [ C ] = spatial_correlation_model(corr_model, X, Y, ros, alpha );
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    %%  topology loop (for each topology, generate a 2D signal )

    for tpl=1:tplMax
        disp(['-- Topology = ', num2str(tpl),'/',num2str(tplMax)]);
        
        %%
        % Generate a synthetic 2D signal using the correlation matrix for T subsequent time slots
        [synt_data, ~, ~, RoT] = generate_synthetic_signal(dim_grid, T, C, r);

        %%
        % Network deployment - place the nodes in the simulation area
        [ S, A, Nei_mat, NeiWithSink, IX, Dist_mat ] = network_deployment( SoF_X, SoF_Y, NoS, synt_data, T, TR  );

        %%
        % calculate the angle of each sensor regard to the center of the area and x axis [line (0,0) and (250,0)]
        [angle_vector]=get_angle_vector(S, SoF_X/2, SoF_Y/2, NoS);
        
        cost_vec = ones(length(A(:,1)),1).*r_s;

        %%
        % number of time steps used for the initial training (i.e., estimation of spatio-temporal features of the signal)
        % during these initial time slots, all the sensors sense and transmit their readings to the sink
        %%
        initial_training = 10;

        %%
        % size of the window of consecutive time steps used in the CB-CS
        W = 10;
                
        for ptx=1:range  % transmission probability
            disp([9,'-- Pro. Trx Loop = ',num2str(ptx),'/',num2str(range)]);

            %%
            % Deterministic Node Selection (DNS)
            set_signals = A(:,1:initial_training);
            cov_matrix=cov(set_signals');
            variances=diag(cov_matrix);
            std_dev=sqrt(variances);
            cov_function=cov_matrix;
            [time_table] = time_table_for_tdma_weight_update_conditional_based(ptx, NoS, T, initial_training, cov_function);
             
            %%
            % Generate clusters for Distributed Source Coding (DSC) and assign sensors to the corresponding clusters
            initial_p_tx = ptx*0.1;
            alpha_cluster=2*(1-initial_p_tx);
            [ Cid, ~, Belonging ] = clustering_dsc( NoS, SoF_X, alpha_cluster, TR, IX, Dist_mat);
            
            %%
            % Compute hop counts from each sensor to the sink
            clustering = false;
            [~, No_of_hops ] = get_hops_vector( NoS, NeiWithSink, Belonging, Nei_mat, Cid, clustering);            

            % <<<-------------------------- Initializations end here.
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Spatio-temporal Compression Algos
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%            
           
            disp([9,9,'CS-RNS']);
            TS_type = 'real';
            [~, tot_cost_rns, real_error_rns] = ...
                BLOCKmain_cs_rns(A, cost_vec, initial_p_tx, 2, No_of_hops, TS_type);
            report_CS_rns(tpl,ptx,1) = mean(real_error_rns(initial_training+1:T));
            report_CS_rns(tpl,ptx,2) = sum(tot_cost_rns);
            
            disp([9,9,'Theoretical']);
            [~, tot_cost, real_error] = BLOCKmain_cs_dns(A, cost_vec, No_of_hops, 'theoretical', r_s, Dist_mat, 'Simple_Cov', initial_training, time_table);
            report_CS_lower_bnd(tpl,ptx,1) = mean(real_error(initial_training+1:T));
            report_CS_lower_bnd(tpl,ptx,2) = tot_cost;
            
            disp([9,9,'CS-DNS']);
            [~, tot_cost_shared, real_error_shared] = BLOCKmain_cs_dns(A,cost_vec, No_of_hops, 'real', r_s, Dist_mat, 'Simple_Cov', initial_training, time_table);
            report_CS_dns(tpl,ptx,1) = mean(real_error_shared(initial_training+1:T));
            report_CS_dns(tpl,ptx,2) = tot_cost_shared;
            
            disp([9,9,'KRON-CS']);
            [~,tot_cost_shared, real_error_shared] = BLOCKmain_CB_CS(A,cost_vec, No_of_hops, 'real_sec', r_s, Dist_mat, 'Simple_Cov', initial_training, time_table, W, 'dct');
            report_KRON_CS(tpl,ptx,1) = mean(real_error_shared(initial_training+1:T));
            report_KRON_CS(tpl,ptx,2) = tot_cost_shared;
            
            disp([9,9,'CB-CS(iid)']);
            [~,tot_cost_shared, real_error_shared] = BLOCKmain_CB_CS(A,cost_vec, No_of_hops, 'real_sec', r_s, Dist_mat, 'Variogram_Est', initial_training, time_table, W, 'iden');
            report_CB_CS_iid(tpl,ptx,1) = mean(real_error_shared(initial_training+1:T));
            report_CB_CS_iid(tpl,ptx,2) = tot_cost_shared;
            
            disp([9,9,'CB-CS(dct)']);
            [~,tot_cost_shared, real_error_shared] = BLOCKmain_CB_CS(A,cost_vec, No_of_hops, 'real_sec', r_s, Dist_mat, 'Variogram_Est', initial_training, time_table, W, 'dct');
            report_CB_CS_dct(tpl,ptx,1) = mean(real_error_shared(initial_training+1:T));
            report_CB_CS_dct(tpl,ptx,2) = tot_cost_shared;
            
            disp([9,9,'DSC']);
            [~,energy_cons_dsc, real_error_dsc] = BLOCKmain_dsc(A, T, No_of_hops, r_s, rate4, Cid, NoS, all_indeces4);
            report_dsc(tpl,ptx,1) = mean(real_error_dsc(initial_training+1:T));
            report_dsc(tpl,ptx,2) = energy_cons_dsc;
            
        end
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Temporal Compression Algos
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        disp([9,9,'LTC']);
        disp([9,9,'DCT']);
                           
        for ptx = 1:range
            tol = squeeze(report_CS_rns(tpl,ptx,1)); % threshold that temporal algorithms consider for the compression
            cr_ltc = zeros(1,NoS);
            rmse_ltc = zeros(1,NoS);
            n_op_ltc = zeros(4,NoS);
            max_error_ltc = zeros(1,NoS);
            
            cr_dct = zeros(1,NoS);
            rmse_dct = zeros(1,NoS);
            n_op_dct = zeros(4,NoS);
            max_error_dct = zeros(1,NoS);
            
            for s = 1:NoS
                time_series = squeeze(synt_data(S(s,1),S(s,2),:))';
                %disp([9,9,'LTC']);
                [cr_ltc(s),rmse_ltc(s),n_op_ltc(:,s),rec_x_ltc,max_error_ltc(s)] = ltc_encoder_main(time_series,tol);
                %disp([9,9,'DCT']);
                [cr_dct(s),rmse_dct(s),n_op_dct(:,s),rec_x_dct,max_error_dct(s)] = dct_encoder_main(time_series,tol);
            end
            
            mean_cr_ltc(tpl,ptx) = mean(cr_ltc);
            mean_energy_ltc(tpl,ptx) = (mean(n_op_ltc,2)'*op_cost)*energy_per_instruction;
            mean_rmse_ltc(tpl,ptx) = mean(rmse_ltc);
            mean_max_error_ltc(tpl,ptx) = mean(max_error_ltc);
            
            mean_cr_dct(tpl,ptx) = mean(cr_dct);
            mean_energy_dct(tpl,ptx) = (mean(n_op_dct,2)'*op_cost)*energy_per_instruction;
            mean_rmse_dct(tpl,ptx) = mean(rmse_dct);
            mean_max_error_dct(tpl,ptx) = mean(max_error_dct);
            
            tx_energy_ltc(tpl,ptx) = energy_consumption_header( No_of_hops', [1:NoS], T*r_s*cr_ltc );
            tx_energy_dct(tpl,ptx) = energy_consumption_header( No_of_hops', [1:NoS], T*r_s*cr_dct );
        end
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% Compute performance metrics (reconstruction error and energy consumption)
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    error_CS_lower_bnd(r_id,:) = mean(report_CS_lower_bnd(:,:,1));
    energy_CS_lower_bnd(r_id,:) = mean(report_CS_lower_bnd(:,:,2));
    
    error_CS_dns(r_id,:) = mean(report_CS_dns(:,:,1));
    energy_CS_dns(r_id,:) = mean(report_CS_dns(:,:,2));
    
    error_KRON_CS(r_id,:) = mean(report_KRON_CS(:,:,1));
    energy_KRON_CS(r_id,:) = mean(report_KRON_CS(:,:,2));
    
    error_CB_CS_iid(r_id,:) = mean(report_CB_CS_iid(:,:,1));
    energy_CB_CS_iid(r_id,:) = mean(report_CB_CS_iid(:,:,2));
    
    error_CB_CS_dct(r_id,:) = mean(report_CB_CS_dct(:,:,1));
    energy_CB_CS_dct(r_id,:) = mean(report_CB_CS_dct(:,:,2));
    
    error_dsc(r_id,:) = mean(report_dsc(:,:,1));
    energy_dsc(r_id,:) = mean(report_dsc(:,:,2));
    
    error_ltc(r_id,:) = mean(mean_max_error_ltc);
    energy_ltc(r_id,:) = mean(mean_energy_ltc).*1e+9 + mean(tx_energy_ltc);
    
    error_dct(r_id,:) = mean(mean_max_error_dct);
    energy_dct(r_id,:) = mean(mean_energy_dct).*1e+9 + mean(tx_energy_dct);

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% Plot and save results
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                                                               
    matrix_err = zeros(10,16);
    h=figure;
    hold on
    plot(squeeze(energy_CS_dns(r_id,:)),squeeze(error_CS_lower_bnd(r_id,:)), 'm','LineWidth',2)
    matrix_err(:,1)=squeeze(energy_CS_dns(r_id,:))';
    matrix_err(:,2)=squeeze(error_CS_lower_bnd(r_id,:))';
    plot(squeeze(energy_CS_dns(r_id,:)),squeeze(error_CS_dns(r_id,:)), 'b','LineWidth',2)
    matrix_err(:,3)=squeeze(energy_CS_dns(r_id,:))';
    matrix_err(:,4)=squeeze(error_CS_dns(r_id,:))';
    plot(squeeze(energy_KRON_CS(r_id,:)),squeeze(error_KRON_CS(r_id,:)), 'g-o','LineWidth',2)
    matrix_err(:,5)=squeeze(energy_KRON_CS(r_id,:))';
    matrix_err(:,6)=squeeze(error_KRON_CS(r_id,:))';
    plot(squeeze(energy_CB_CS_iid(r_id,:)),squeeze(error_CB_CS_iid(r_id,:)), 'r-o','LineWidth',2)
    matrix_err(:,7)=squeeze(energy_CB_CS_iid(r_id,:))';
    matrix_err(:,8)=squeeze(error_CB_CS_iid(r_id,:))';
    plot(squeeze(energy_CB_CS_dct(r_id,:)),squeeze(error_CB_CS_dct(r_id,:)), 'b-o','LineWidth',2)
    matrix_err(:,9)=squeeze(energy_CB_CS_dct(r_id,:))';
    matrix_err(:,10)=squeeze(error_CB_CS_dct(r_id,:))';
    plot(squeeze(energy_dsc(r_id,:)),squeeze(error_dsc(r_id,:)), 'b--','LineWidth',2)
    matrix_err(:,11)=squeeze(energy_dsc(r_id,:))';
    matrix_err(:,12)=squeeze(error_dsc(r_id,:))';
    plot(squeeze(energy_ltc(r_id,:)),squeeze(error_ltc(r_id,:)), 'r--','LineWidth',2)
    matrix_err(:,13)=squeeze(energy_ltc(r_id,:))';
    matrix_err(:,14)=squeeze(error_ltc(r_id,:))';
    plot(squeeze(energy_dct(r_id,:)),squeeze(error_dct(r_id,:)), 'g--','LineWidth',2)
    matrix_err(:,15)=squeeze(energy_dct(r_id,:))';
    matrix_err(:,16)=squeeze(error_dct(r_id,:))';
    
    legend( 'Lower Bound', 'CS-DNS', 'Kron-CS', 'CB-CS (iid)', 'CB-CS (DCT)', 'DSC', 'LTC', 'DCT')
    xlabel('Energy consumption (nJ)')
    ylabel('Error $\hat{x}-x$', 'Interpreter', 'Latex');
    titlename = strcat(' Time instants=', num2str(T), ' | ', num2str(NoS),' nodes | ', num2str(SoF_X), 'x', num2str(SoF_Y), ' grid | Tx range=', num2str(TR), ' | rho_t=', num2str(r), ' | rho_s=', num2str(ros));
    title(titlename);
    name=strcat('erro_rho_t', num2str(r*100), 'rho_s', num2str(ros*1000));
    t_name=strcat(name,'.txt');
    f = fullfile('Results',t_name);
    f2 = fullfile('Results',name);
    dlmwrite(f, matrix_err, '\t');
    saveas(h, f2,'fig')
    saveas(h, f2,'jpg')
    
end
close all
display('Simulation finished.');